@@ -56,8 +56,8 @@ get_server_information(struct dbus *bus, DBusMessage *msg)
     if (!dbus_connection_send(bus->conn, reply, NULL))
         goto err;
 
-    if (dbus_connection_has_messages_to_send(bus->conn))
-        fdm_event_add(bus->fdm, bus->bus_fd, EPOLLOUT);
+    dbus_connection_flush(bus->conn);
+    assert(!dbus_connection_has_messages_to_send(bus->conn));
     ret = true;
 
 err:
@@ -81,13 +81,14 @@ get_capabilities(struct dbus *bus, DBusMessage *msg)
     dbus_message_iter_append_basic(&arr, DBUS_TYPE_STRING, &(const char *){"body-markup"});
     dbus_message_iter_append_basic(&arr, DBUS_TYPE_STRING, &(const char *){"actions"});
     dbus_message_iter_append_basic(&arr, DBUS_TYPE_STRING, &(const char *){"icon-static"});
+    dbus_message_iter_append_basic(&arr, DBUS_TYPE_STRING, &(const char *){"x-canonical-private-synchronous"});
 
     dbus_message_iter_close_container(&iter, &arr);
     if (!dbus_connection_send(bus->conn, reply, NULL))
         goto err;
 
-    if (dbus_connection_has_messages_to_send(bus->conn))
-        fdm_event_add(bus->fdm, bus->bus_fd, EPOLLOUT);
+    dbus_connection_flush(bus->conn);
+    assert(!dbus_connection_has_messages_to_send(bus->conn));
     ret = true;
 
 err:
@@ -121,7 +122,7 @@ notify(struct dbus *bus, DBusMessage *msg)
     if (notif_mgr_is_paused(bus->notif_mgr)) {
         LOG_DBG("paused, refusing to notify");
     } else {
-        char *app_name, *app_icon, *summary, *body;
+        char *app_name, *app_icon, *summary, *body, *sync_tag = NULL;
         enum urgency urgency = URGENCY_NORMAL;
         int8_t progress_percent = -1;
 
@@ -253,8 +254,6 @@ notify(struct dbus *bus, DBusMessage *msg)
             dbus_message_iter_recurse(&entry_iter, &value_iter);
             dbus_message_iter_next(&entry_iter);
 
-            LOG_DBG("hint: %s", name);
-
             if (strcmp(name, "urgency") == 0) {
                 if (dbus_message_iter_get_arg_type(&value_iter) != DBUS_TYPE_BYTE)
                     goto err;
@@ -267,6 +266,14 @@ notify(struct dbus *bus, DBusMessage *msg)
                 urgency = level;
             }
 
+            else if (strcmp(name, "x-canonical-private-synchronous") == 0) {
+                if (dbus_message_iter_get_arg_type(&value_iter) != DBUS_TYPE_STRING)
+                    goto err;
+
+                dbus_message_iter_get_basic(&value_iter, &sync_tag);
+                LOG_DBG("x-canonical-private-synchronous: %s", sync_tag);
+            }
+
             else if (strcmp(name, "value") == 0) {
                 if (dbus_message_iter_get_arg_type(&value_iter) != DBUS_TYPE_INT32)
                     goto err;
@@ -298,13 +305,17 @@ notify(struct dbus *bus, DBusMessage *msg)
                     image_path = path;
                 }
 
-                if (pix != NULL) {
-                    free(pixman_image_get_data(pix));
-                    pixman_image_unref(pix);
-                    pix = NULL;
-                }
+                pixman_image_t *image = icon_load(
+                    image_path, bus->conf->max_icon_size, bus->icon_theme);
 
-                pix = icon_load(image_path, bus->conf->max_icon_size, bus->icon_theme);
+                if (image != NULL) {
+                    if (pix != NULL) {
+                        free(pixman_image_get_data(pix));
+                        pixman_image_unref(pix);
+                    }
+
+                    pix = image;
+                }
 
                 free(scheme);
                 free(host);
@@ -397,9 +408,38 @@ notify(struct dbus *bus, DBusMessage *msg)
                         pix = NULL;
                     }
 
+                    /* pixman expects pre-multiplied alpha */
+                    if (format == PIXMAN_a8b8g8r8) {
+                        for (int i = 0; i < height; i++) {
+                            uint32_t *p = (uint32_t *)&image_data[i * stride];
+                            for (int j = 0; j < width; j++, p++) {
+                                uint8_t a = (*p >> 24) & 0xff;
+                                uint8_t b = (*p >> 16) & 0xff;
+                                uint8_t g = (*p >> 8) & 0xff;
+                                uint8_t r = (*p >> 0) & 0xff;
+
+                                if (a == 0xff)
+                                    continue;
+
+                                if (a == 0) {
+                                    r = g = b = 0;
+                                } else {
+                                    r = r * a / 0xff;
+                                    g = g * a / 0xff;
+                                    b = b * a / 0xff;
+                                }
+
+                                *p = (uint32_t)a << 24 | b << 16 | g << 8 | r;
+                            }
+                        }
+                    }
+
                     pix = pixman_image_create_bits_no_clear(
                         format, width, height, (uint32_t *)image_data, stride);
                 }
+            } else {
+                LOG_DBG("hint: %s unrecognized, ignoring", name);
+
             }
         }
 
@@ -412,7 +452,7 @@ notify(struct dbus *bus, DBusMessage *msg)
         dbus_message_iter_get_basic(&args_iter, &timeout_ms);
         LOG_DBG("timeout = %dms", timeout_ms);
 
-        notif = notif_mgr_create_notif(bus->notif_mgr, replaces_id);
+        notif = notif_mgr_create_notif(bus->notif_mgr, replaces_id, sync_tag);
         if (notif == NULL)
             goto err;
 
@@ -448,8 +488,8 @@ notify(struct dbus *bus, DBusMessage *msg)
     if (!dbus_connection_send(bus->conn, reply, NULL))
         goto err;
 
-    if (dbus_connection_has_messages_to_send(bus->conn))
-        fdm_event_add(bus->fdm, bus->bus_fd, EPOLLOUT);
+    dbus_connection_flush(bus->conn);
+    assert(!dbus_connection_has_messages_to_send(bus->conn));
     ret = true;
     goto out;
 
@@ -506,8 +546,84 @@ close_notification(struct dbus *bus, DBusMessage *msg)
     if (!dbus_connection_send(bus->conn, reply, NULL))
         goto err;
 
-    if (dbus_connection_has_messages_to_send(bus->conn))
-        fdm_event_add(bus->fdm, bus->bus_fd, EPOLLOUT);
+    dbus_connection_flush(bus->conn);
+    assert(!dbus_connection_has_messages_to_send(bus->conn));
+    ret = true;
+
+err:
+    dbus_message_unref(reply);
+    return ret;
+}
+
+static bool
+introspect(struct dbus *bus, DBusMessage *msg)
+{
+    const char *data =
+        "<!DOCTYPE node PUBLIC \"-//freedesktop//DTD D-BUS Object Introspection 1.0//EN\"\n"
+        " \"http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd\">\n"
+        "<node name=\"/org/freedesktop/Notifications\">\n"
+        "  <interface name=\"org.freedesktop.Notifications\">\n"
+
+        "    <method name=\"Notify\">\n"
+        "      <arg name=\"id\" type=\"u\" direction=\"out\"/>\n"
+        "      <arg name=\"app_name\" type=\"s\" direction=\"in\"/>\n"
+        "      <arg name=\"replaces_id\" type=\"u\" direction=\"in\"/>\n"
+        "      <arg name=\"app_icon\" type=\"s\" direction=\"in\"/>\n"
+        "      <arg name=\"summary\" type=\"s\" direction=\"in\"/>\n"
+        "      <arg name=\"body\" type=\"s\" direction=\"in\"/>\n"
+        "      <arg name=\"actions\" type=\"as\" direction=\"in\"/>\n"
+        "      <arg name=\"hints\" type=\"a{sv}\" direction=\"in\"/>\n"
+        "      <arg name=\"expire_timeout\" type=\"i\" direction=\"in\"/>\n"
+        "    </method>\n"
+
+        "    <method name=\"CloseNotification\">\n"
+        "      <arg name=\"id\" type=\"u\" direction=\"in\"/>\n"
+        "    </method>\n"
+
+        "    <method name=\"GetServerInformation\">\n"
+        "      <arg name=\"name\" type=\"s\" direction=\"out\"/>\n"
+        "      <arg name=\"vendor\" type=\"s\" direction=\"out\"/>\n"
+        "      <arg name=\"version\" type=\"s\" direction=\"out\"/>\n"
+        "      <arg name=\"spec_version\" type=\"s\" direction=\"out\"/>\n"
+        "    </method>\n"
+
+        "    <method name=\"GetCapabilities\">\n"
+        "      <arg name=\"capabilities\" type=\"as\" direction=\"out\"/>\n"
+        "    </method>\n"
+
+        "    <signal name=\"NotificationClosed\">\n"
+        "      <arg name=\"id\" type=\"u\"/>\n"
+        "      <arg name=\"reason\" type=\"u\"/>\n"
+        "    </signal>\n"
+
+        "    <signal name=\"ActionInvoked\">\n"
+        "      <arg name=\"id\" type=\"u\"/>\n"
+        "      <arg name=\"action_key\" type=\"s\"/>\n"
+        "    </signal>\n"
+
+        "    <signal name=\"ActivationToken\">\n"
+        "      <arg name=\"id\" type=\"u\"/>\n"
+        "      <arg name=\"activation_token\" type=\"s\"/>\n"
+        "    </signal>\n"
+
+        "  </interface>\n"
+        "</node>\n";
+
+    bool ret = false;
+    DBusMessage *reply = dbus_message_new_method_return(msg);
+
+    if (reply == NULL)
+        goto err;
+
+    DBusMessageIter args;
+    dbus_message_iter_init_append(reply, &args);
+    dbus_message_iter_append_basic(&args, DBUS_TYPE_STRING, &data);
+
+    if (!dbus_connection_send(bus->conn, reply, NULL))
+        goto err;
+
+    dbus_connection_flush(bus->conn);
+    assert(!dbus_connection_has_messages_to_send(bus->conn));
     ret = true;
 
 err:
@@ -520,21 +636,27 @@ dbus_handler(DBusConnection *conn, DBusMessage *msg, void *data)
 {
     struct dbus *bus = data;
 
-    const char *iface __attribute__((unused)) = dbus_message_get_interface(msg);
+    const char *iface = dbus_message_get_interface(msg);
     const char *member = dbus_message_get_member(msg);
     LOG_DBG("%s:%s", iface, member);
 
     static const struct {
+        const char *iface;
         const char *name;
         bool (*handler)(struct dbus *bus, DBusMessage *msg);
     } handlers[] = {
-        {"GetServerInformation", &get_server_information},
-        {"GetCapabilities", &get_capabilities},
-        {"Notify", &notify},
-        {"CloseNotification", &close_notification},
+        /* Don't forget to update introspect() when adding methods or signals */
+        {"org.freedesktop.DBus.Introspectable", "Introspect", &introspect},
+
+        {"org.freedesktop.Notifications", "GetServerInformation", &get_server_information},
+        {"org.freedesktop.Notifications", "GetCapabilities", &get_capabilities},
+        {"org.freedesktop.Notifications", "Notify", &notify},
+        {"org.freedesktop.Notifications", "CloseNotification", &close_notification},
     };
 
     for (size_t i = 0; i < sizeof(handlers) / sizeof(handlers[0]); i++) {
+        if (strcmp(handlers[i].iface, iface) != 0)
+            continue;
         if (strcmp(handlers[i].name, member) != 0)
             continue;
 
@@ -562,28 +684,50 @@ signal_notification_closed(struct dbus *bus, uint32_t id, uint32_t reason)
     dbus_message_iter_append_basic(&iter, DBUS_TYPE_UINT32, &(dbus_uint32_t){id});
     dbus_message_iter_append_basic(&iter, DBUS_TYPE_UINT32, &(dbus_uint32_t){reason});
     dbus_connection_send(bus->conn, signal, NULL);
+    dbus_connection_flush(bus->conn);
+    assert(!dbus_connection_has_messages_to_send(bus->conn));
     dbus_message_unref(signal);
-    if (dbus_connection_has_messages_to_send(bus->conn))
-        fdm_event_add(bus->fdm, bus->bus_fd, EPOLLOUT);
     return true;
 }
 
 bool
 dbus_signal_expired(struct dbus *bus, uint32_t id)
 {
-    return signal_notification_closed(bus, id, 0);
+    return signal_notification_closed(bus, id, 1);
 }
 
 bool
 dbus_signal_dismissed(struct dbus *bus, uint32_t id)
 {
-    return signal_notification_closed(bus, id, 1);
+    return signal_notification_closed(bus, id, 2);
 }
 
 bool
 dbus_signal_closed(struct dbus *bus, uint32_t id)
 {
-    return signal_notification_closed(bus, id, 2);
+    return signal_notification_closed(bus, id, 3);
+}
+
+bool
+dbus_signal_token(struct dbus *bus, uint32_t id, const char *token)
+{
+    DBusMessage *signal = dbus_message_new_signal(
+        "/org/freedesktop/Notifications",
+        "org.freedesktop.Notifications",
+        "ActivationToken");
+
+    if (signal == NULL)
+        return false;
+
+    DBusMessageIter iter;
+    dbus_message_iter_init_append(signal, &iter);
+    dbus_message_iter_append_basic(&iter, DBUS_TYPE_UINT32, &(dbus_uint32_t){id});
+    dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &token);
+    dbus_connection_send(bus->conn, signal, NULL);
+    dbus_connection_flush(bus->conn);
+    assert(!dbus_connection_has_messages_to_send(bus->conn));
+    dbus_message_unref(signal);
+    return true;
 }
 
 bool
@@ -602,9 +746,9 @@ dbus_signal_action(struct dbus *bus, uint32_t id, const char *action_id)
     dbus_message_iter_append_basic(&iter, DBUS_TYPE_UINT32, &(dbus_uint32_t){id});
     dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &action_id);
     dbus_connection_send(bus->conn, signal, NULL);
+    dbus_connection_flush(bus->conn);
+    assert(!dbus_connection_has_messages_to_send(bus->conn));
     dbus_message_unref(signal);
-    if (dbus_connection_has_messages_to_send(bus->conn))
-        fdm_event_add(bus->fdm, bus->bus_fd, EPOLLOUT);
     return true;
 }
 
@@ -614,23 +758,11 @@ fdm_handler(struct fdm *fdm, int fd, int events, void *data)
     bool ret = false;
     struct dbus *bus = data;
 
-    bool had_outgoing = dbus_connection_has_messages_to_send(bus->conn);
-
-    /* A sending function added EPOLLOUT when it wasn't necessary */
-    if ((events & EPOLLOUT) && !had_outgoing) {
-        LOG_WARN("EPOLLOUT set, but no outgoing messages");
-        fdm_event_del(bus->fdm, bus->bus_fd, EPOLLOUT);
-    }
-
     if (!dbus_connection_read_write(bus->conn, 0)) {
         LOG_ERRNO("failed to read/write dbus connection");
         goto err;
     }
 
-    /* Remove EPOLLOUT when no longer needed */
-    if (had_outgoing && !dbus_connection_has_messages_to_send(bus->conn))
-        fdm_event_del(bus->fdm, bus->bus_fd, EPOLLOUT);
-
     while (dbus_connection_dispatch(bus->conn) != DBUS_DISPATCH_COMPLETE)
         ;
 
