@@ -76,11 +76,12 @@ struct notif {
     bool is_configured;
 
     uint32_t id;
-    enum urgency urgency;
+    char *synchronous_tag;  /* x-canonical-private-synchronous */
 
     char32_t *app;
     char32_t *summary;
     char32_t *body;
+    enum urgency urgency;
     tll(struct action) actions;
 
     int8_t progress;
@@ -180,6 +181,8 @@ notif_mgr_destroy(struct notif_mgr *mgr)
 
     regfree(&mgr->html_entity_re);
 
+    notif_mgr_dismiss_all(mgr);
+
     tll_foreach(mgr->notifs, it)
         notif_destroy(it->item);
     tll_free(mgr->notifs);
@@ -400,6 +403,23 @@ notif_mgr_get_notif(struct notif_mgr *mgr, uint32_t id)
     return NULL;
 }
 
+struct notif *
+notif_mgr_get_notif_for_sync_tag(struct notif_mgr *mgr, const char *tag)
+{
+    if (tag == NULL)
+        return NULL;
+
+    tll_foreach(mgr->notifs, it) {
+        if (it->item->synchronous_tag == NULL)
+            continue;
+
+        if (strcmp(it->item->synchronous_tag, tag) == 0)
+            return it->item;
+    }
+
+    return NULL;
+}
+
 struct notif *
 notif_mgr_get_notif_for_surface(struct notif_mgr *mgr,
                                 const struct wl_surface *surface)
@@ -415,10 +435,17 @@ notif_mgr_get_notif_for_surface(struct notif_mgr *mgr,
 /* Instantiates a new notification. You *must* call
  * notif_mgr_refresh() "soon" (after configuring the notification). */
 struct notif *
-notif_mgr_create_notif(struct notif_mgr *mgr, uint32_t replaces_id)
+notif_mgr_create_notif(struct notif_mgr *mgr, uint32_t replaces_id,
+                       const char *sync_tag)
 {
     int notif_id;
 
+    {
+        struct notif *old_notif = notif_mgr_get_notif_for_sync_tag(mgr, sync_tag);
+        if (old_notif != NULL)
+            return old_notif;
+    }
+
     if (replaces_id != 0) {
         struct notif *old_notif = notif_mgr_get_notif(mgr, replaces_id);
         if (old_notif != NULL)
@@ -432,10 +459,11 @@ notif_mgr_create_notif(struct notif_mgr *mgr, uint32_t replaces_id)
     *notif = (struct notif) {
         .mgr = mgr,
         .id = notif_id,
-        .urgency = URGENCY_NORMAL,
+        .synchronous_tag = sync_tag != NULL ? strdup(sync_tag) : NULL,
         .app = c32dup(U""),
         .summary = c32dup(U""),
         .body = c32dup(U""),
+        .urgency = URGENCY_NORMAL,
         .actions = tll_init(),
         .timeout_ms = -1,  /* -1, up to us, 0 - never expire */
         .timeout_fd = -1,
@@ -529,6 +557,7 @@ notif_destroy(struct notif *notif)
         tll_remove(notif->text_run_cache, it);
     }
 
+    free(notif->synchronous_tag);
     free(notif->app);
     free(notif->summary);
     free(notif->body);
@@ -1712,6 +1741,33 @@ expand_format_string(const struct notif *notif, const char32_t *fmt)
     return ret;
 }
 
+static inline void
+fill_rounded_rectangle(pixman_op_t op, pixman_image_t* dest,
+                       const pixman_color_t* color, int16_t x, int16_t y,
+                       uint16_t width, uint16_t height, uint16_t radius)
+{
+    int rect_count = ( radius + radius ) + 1;
+    pixman_rectangle16_t rects[rect_count];
+
+    for (int i = 0; i <= radius; i++){
+        uint16_t ydist = radius - i;
+        uint16_t curve = sqrt(radius * radius - ydist * ydist);
+
+        rects[i] = (pixman_rectangle16_t){
+            x + radius - curve, y + i, width - radius * 2 + curve * 2, 1
+        };
+
+        rects[radius + i] = (pixman_rectangle16_t){
+            x + radius - curve, y + height - i, width - radius * 2 + curve * 2, 1
+        };
+    }
+
+    rects[(radius * 2)] = (pixman_rectangle16_t){
+        x, y + radius, width, height + 1 - radius * 2
+    };
+    pixman_image_fill_rectangles(op, dest, color,rect_count, rects);
+}
+
 static int
 notif_show(struct notif *notif, int y)
 {
@@ -1881,30 +1937,71 @@ notif_show(struct notif *notif, int y)
         roundf(conf->margins.horizontal / scale));  /* left */
 
     struct buffer *buf = wayl_get_buffer(wayl, width, height);
-    const int brd_sz = urgency->border.size;;
+    const int brd_sz = urgency->border.size;
+    const int brd_rad = min(min(urgency->border.radius, buf->width*0.5), buf->height*0.5);
 
     pixman_region32_t clip;
     pixman_region32_init_rect(&clip, 0, 0, width, height);
     pixman_image_set_clip_region32(buf->pix, &clip);
     pixman_region32_fini(&clip);
 
-    /* Border */
-    pixman_image_fill_rectangles(
-        PIXMAN_OP_SRC, buf->pix, &urgency->border.color,
-        4, (pixman_rectangle16_t []){
-            {0, 0, buf->width, brd_sz},                     /* top */
-            {buf->width - brd_sz, 0, brd_sz, buf->height},  /* right */
-            {0, buf->height - brd_sz, buf->width, brd_sz},  /* bottom */
-            {0, 0, brd_sz, buf->height},                    /* left */
-    });
+    if (brd_rad == 0){
+        /* Border */
+        pixman_image_fill_rectangles(
+            PIXMAN_OP_SRC, buf->pix, &urgency->border.color,
+            4, (pixman_rectangle16_t []){
+                {0, 0, buf->width, brd_sz},                     /* top */
+                {buf->width - brd_sz, 0, brd_sz, buf->height},  /* right */
+                {0, buf->height - brd_sz, buf->width, brd_sz},  /* bottom */
+                {0, 0, brd_sz, buf->height},                    /* left */
+            });
 
-    /* Background */
-    pixman_image_fill_rectangles(
-        PIXMAN_OP_SRC, buf->pix, &urgency->bg,
-        1, &(pixman_rectangle16_t){
-            brd_sz, brd_sz,
-            buf->width - 2 * brd_sz, buf->height - 2 * brd_sz}
-    );
+        /* Background */
+        pixman_image_fill_rectangles(
+            PIXMAN_OP_SRC, buf->pix, &urgency->bg,
+            1, &(pixman_rectangle16_t){
+                brd_sz, brd_sz,
+                buf->width - 2 * brd_sz, buf->height - 2 * brd_sz}
+          );
+    } else {
+        const int msaa_scale = 2;
+        const double brd_sz_scaled = brd_sz * msaa_scale;
+        const double brd_rad_scaled = brd_rad * msaa_scale;
+        int w = buf->width * msaa_scale;
+        int h = buf->height * msaa_scale;
+
+        pixman_image_t *bg;
+        if (msaa_scale != 1){
+            bg = pixman_image_create_bits(PIXMAN_a8r8g8b8, w, h, NULL, w*4);
+        } else {
+            bg = buf->pix;
+        }
+
+        /* Border */
+        fill_rounded_rectangle(
+            PIXMAN_OP_SRC, bg, &urgency->border.color,
+            0, 0, w, h, brd_rad_scaled);
+
+        /* Background */
+        fill_rounded_rectangle(
+            PIXMAN_OP_SRC, bg, &urgency->bg, brd_sz_scaled, brd_sz_scaled,
+            w-(brd_sz_scaled*2),
+            h-(brd_sz_scaled*2),
+            brd_rad_scaled - brd_sz_scaled);
+
+        if (msaa_scale != 1){
+            pixman_f_transform_t ftrans;
+            pixman_transform_t trans;
+            pixman_f_transform_init_scale(&ftrans, msaa_scale, msaa_scale);
+            pixman_transform_from_pixman_f_transform(&trans, &ftrans);
+            pixman_image_set_transform(bg, &trans);
+            pixman_image_set_filter(bg, PIXMAN_FILTER_BILINEAR, NULL, 0);
+
+            pixman_image_composite32(
+                PIXMAN_OP_SRC, bg, NULL, buf->pix, 0, 0, 0, 0, 0, 0, buf->width, buf->height);
+            pixman_image_unref(bg);
+        }
+    }
 
     /* Image */
     if (notif->pix != NULL) {
@@ -2273,6 +2370,9 @@ fdm_action_writer(struct fdm *fdm, int fd, int events, void *data)
         async->input_len - async->input_idx);
 
     if (count < 0) {
+        if (errno == EINTR)
+            return true;
+
         LOG_ERRNO("could not write actions to actions selection helper");
         goto done;
     }
@@ -2297,25 +2397,33 @@ fdm_action_reader(struct fdm *fdm, int fd, int events, void *data)
 {
     struct action_async *async = data;
 
-    const size_t chunk_sz = 128;
+    const size_t chunk_sz = 1024;
     char buf[chunk_sz];
 
     ssize_t count = read(async->from_child, buf, chunk_sz);
     if (count < 0) {
+        if (errno == EINTR)
+            return true;
+
         LOG_ERRNO("failed to read from actions selection helper");
-        goto check_pollhup;
+
+        /* FIXME: leaks memory (the async context) */
+        return false;
     }
 
-    /* Append to previously received response */
-    size_t new_len = async->output_len + count;
-    async->output = realloc(async->output, new_len);
-    memcpy(&async->output[async->output_len], buf, count);
-    async->output_len = new_len;
+    if (count > 0) {
+        /* Append to previously received response */
+        size_t new_len = async->output_len + count;
+        async->output = realloc(async->output, new_len);
+        memcpy(&async->output[async->output_len], buf, count);
+        async->output_len = new_len;
 
-check_pollhup:
-
-    if (!(events & EPOLLHUP))
+        /* There may be more data to read */
         return true;
+    }
+
+    /* No more data to read */
+    assert(count == 0);
 
     /* Strip trailing spaces/newlines */
     while (async->output_len > 0
@@ -2382,7 +2490,7 @@ check_pollhup:
     LOG_WARN("could not map chosen action label to action ID: %.*s", (int)chosen_len, chosen);
 
 done:
-    completion_cb(notif_id, action_id, cb_data);
+    completion_cb(notif, action_id, cb_data);
     free(chosen);
 
     if (notif->deferred_expiral == EXPIRE_DELAYED) {
@@ -2485,6 +2593,28 @@ notif_action_count(const struct notif *notif)
     return tll_length(notif->actions);
 }
 
+bool
+notif_signal_action(const struct notif *notif, const char *action_id)
+{
+    if (action_id == NULL)
+        return false;
+
+    tll_foreach(notif->actions, it) {
+        if (strcmp(it->item.id, action_id) == 0) {
+            char *activation_token =
+                wayl_get_activation_token(notif->mgr->wayl, notif->surface);
+
+            if (activation_token != NULL)
+                dbus_signal_token(notif->mgr->bus, notif->id, activation_token);
+            free(activation_token);
+
+            return dbus_signal_action(notif->mgr->bus, notif->id, action_id);
+        }
+    }
+
+    return false;
+}
+
 void
 notif_select_action(
     struct notif *notif, notif_select_action_cb completion_cb, void *data)
@@ -2625,7 +2755,7 @@ err_before_fork:
     free(copy);
     free(argv);
 
-    completion_cb(notif->id, NULL, data);
+    completion_cb(notif, NULL, data);
     notif->deferred_dismissal = DISMISS_IMMEDIATELY;
     notif->deferred_expiral = EXPIRE_IMMEDIATELY;
     return;
@@ -2635,7 +2765,7 @@ err_in_parent:
     free(input);
     fdm_del(notif->mgr->fdm, to_child[1]);
     fdm_del(notif->mgr->fdm, from_child[0]);
-    completion_cb(notif->id, NULL, data);
+    completion_cb(notif, NULL, data);
     notif->deferred_dismissal = DISMISS_IMMEDIATELY;
     notif->deferred_expiral = EXPIRE_IMMEDIATELY;
     return;
